=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2006-02-22 05:58:46 +0000
+++ bzrlib/fetch.py	2006-02-28 01:13:01 +0000
@@ -36,6 +36,7 @@
                            MissingText)
 from bzrlib.trace import mutter
 from bzrlib.progress import ProgressBar
+from bzrlib.reconcile import RepoReconciler
 from bzrlib.revision import NULL_REVISION
 from bzrlib.symbol_versioning import *
 
@@ -150,6 +151,14 @@
             self.from_repository.revision_store,
             revs,
             pb=self.pb)
+        # fixup inventory if needed:
+        # this is expensive because we have no inverse index to current ghosts.
+        # but on local disk its a few seconds and sftp push is already insane.
+        # so we just-do-it.
+        # FIXME: the generic code path should not need this, if it truely is
+        # generic.
+        reconciler = RepoReconciler(self.to_repository)
+        reconciler.reconcile()
 
     def _fetch_weave_texts(self, revs):
         file_ids = self.from_repository.fileid_involved_by_set(revs)
@@ -180,23 +189,24 @@
     def _fetch_inventory_weave(self, revs):
         self.pb.update("inventory fetch", 0, 2)
         from_weave = self.from_repository.get_inventory_weave()
-        to_weave = self.to_repository.get_inventory_weave()
+        self.to_inventory_weave = self.to_repository.get_inventory_weave()
         self.pb.update("inventory fetch", 1, 2)
-        to_weave = self.to_control.get_weave('inventory',
+        self.to_inventory_weave = self.to_control.get_weave('inventory',
                 self.to_repository.get_transaction())
         self.pb.update("inventory fetch", 2, 2)
 
-        if to_weave.numversions() > 0:
+        if self.to_inventory_weave.numversions() > 0:
             # destination has contents, must merge
             try:
-                to_weave.join(from_weave, pb=self.pb, msg='merge inventory')
+                self.to_inventory_weave.join(from_weave, pb=self.pb, msg='merge inventory')
             except errors.WeaveParentMismatch:
-                to_weave.reweave(from_weave, pb=self.pb, msg='reweave inventory')
+                self.to_inventory_weave.reweave(from_weave, pb=self.pb, msg='reweave inventory')
         else:
             # destination is empty, just replace it
-            to_weave = from_weave.copy()
+            self.to_inventory_weave = from_weave.copy()
 
-        self.to_control.put_weave('inventory', to_weave,
+        # must be written before pulling any revisions
+        self.to_control.put_weave('inventory', self.to_inventory_weave,
             self.to_repository.get_transaction())
 
         self.pb.clear()

=== modified file 'bzrlib/reconcile.py'
--- bzrlib/reconcile.py	2006-02-26 07:54:02 +0000
+++ bzrlib/reconcile.py	2006-02-28 01:13:01 +0000
@@ -17,6 +17,9 @@
 """Reconcilers are able to fix some potential data errors in a branch."""
 
 
+__all__ = ['reconcile', 'Reconciler', 'RepoReconciler']
+
+
 import bzrlib.branch
 import bzrlib.errors as errors
 import bzrlib.progress
@@ -40,11 +43,7 @@
 
 
 class Reconciler(object):
-    """Reconcilers are used to reconcile existing data.
-
-    Currently this is limited to a single repository, and consists
-    of an inventory reweave with revision cross-checks.
-    """
+    """Reconcilers are used to reconcile existing data."""
 
     def __init__(self, dir):
         self.bzrdir = dir
@@ -60,14 +59,39 @@
         """
         self.pb = ui.ui_factory.progress_bar()
         self.repo = self.bzrdir.find_repository()
+        self.pb.note('Reconciling repository %s',
+                     self.repo.bzrdir.root_transport.base)
+        repo_reconciler = RepoReconciler(self.repo)
+        repo_reconciler.reconcile()
+        self.inconsistent_parents = repo_reconciler.inconsistent_parents
+        self.garbage_inventories = repo_reconciler.garbage_inventories
+        self.pb.note('Reconciliation complete.')
+
+
+class RepoReconciler(object):
+    """Reconciler that reconciles a repository.
+
+    Currently this consists of an inventory reweave with revision cross-checks.
+    """
+
+    def __init__(self, repo):
+        self.repo = repo
+
+    def reconcile(self):
+        """Perform reconciliation.
+        
+        After reconciliation the following attributes document found issues:
+        inconsistent_parents: The number of revisions in the repository whose
+                              ancestry was being reported incorrectly.
+        garbage_inventories: The number of inventory objects without revisions
+                             that were garbage collected.
+        """
+        self.pb = ui.ui_factory.progress_bar()
         self.repo.lock_write()
         try:
-            self.pb.note('Reconciling repository %s',
-                         self.repo.bzrdir.root_transport.base)
             self._reweave_inventory()
         finally:
             self.repo.unlock()
-        self.pb.note('Reconciliation complete.')
 
     def _reweave_inventory(self):
         """Regenerate the inventory weave for the repository from scratch."""
@@ -137,7 +161,7 @@
         assert rev.revision_id == rev_id
         parents = []
         for parent in rev.parent_ids:
-            if parent in self.inventory:
+            if self._parent_is_available(parent):
                 parents.append(parent)
             else:
                 mutter('found ghost %s', parent)
@@ -145,6 +169,15 @@
         if set(self.inventory.parent_names(rev_id)) != set(parents):
             self.inconsistent_parents += 1
 
+    def _parent_is_available(self, parent):
+        """True if parent is a fully available revision
+
+        A fully available revision has a inventory and a revision object in the
+        repository.
+        """
+        return (parent in self._rev_graph or 
+                (parent in self.inventory and self.repo.has_revision(parent)))
+
     def _reweave_step(self, message):
         """Mark a single step of regeneration complete."""
         self.pb.update(message, self.count, self.total)

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2006-02-26 07:54:02 +0000
+++ bzrlib/repository.py	2006-02-28 01:13:01 +0000
@@ -270,15 +270,24 @@
         # revisions. For knits when they are introduced we will probably want
         # to ensure that caching write transactions are in use.
         inv = self.get_inventory_weave()
-        weave_parents = inv.parent_names(revision_id)
-        weave_names = inv.names()
-        for parent_id in r.parent_ids:
+        self._check_revision_parents(r, inv)
+        return r
+
+    def _check_revision_parents(self, revision, inventory):
+        """Private to Repository and Fetch.
+        
+        This checks the parentage of revision in an inventory weave for 
+        consistency and is only applicable to inventory-weave-for-ancestry
+        using repository formats & fetchers.
+        """
+        weave_parents = inventory.parent_names(revision.revision_id)
+        weave_names = inventory.names()
+        for parent_id in revision.parent_ids:
             if parent_id in weave_names:
                 # this parent must not be a ghost.
                 if not parent_id in weave_parents:
                     # but it is a ghost
                     raise errors.CorruptRepository(self)
-        return r
 
     @needs_read_lock
     def get_revision_sha1(self, revision_id):

=== modified file 'bzrlib/tests/interrepository_implementations/test_interrepository.py'
--- bzrlib/tests/interrepository_implementations/test_interrepository.py	2006-02-21 12:42:57 +0000
+++ bzrlib/tests/interrepository_implementations/test_interrepository.py	2006-02-28 01:13:01 +0000
@@ -168,3 +168,46 @@
         to_signature = to_repo.revision_store.get('rev2', 'sig').read()
         self.assertEqual(from_signature, to_signature)
 
+
+class TestCaseWithGhosts(TestCaseWithInterRepository):
+
+    def setUp(self):
+        super(TestCaseWithGhosts, self).setUp()
+        # we want two repositories at this point
+        # one with a revision that is a ghost in the other
+        # repository.
+
+        # 'ghost' is a ghost in missing_ghost and not in with_ghost_rev
+        inv = bzrlib.tree.EmptyTree().inventory
+        repo = self.make_repository('with_ghost_rev')
+        sha1 = repo.add_inventory('ghost', inv, [])
+        rev = bzrlib.revision.Revision(timestamp=0,
+                                       timezone=None,
+                                       committer="Foo Bar <foo@example.com>",
+                                       message="Message",
+                                       inventory_sha1=sha1,
+                                       revision_id='ghost')
+        rev.parent_ids = []
+        repo.add_revision('ghost', rev)
+         
+        repo = self.make_to_repository('missing_ghost')
+        sha1 = repo.add_inventory('with_ghost', inv, [])
+        rev = bzrlib.revision.Revision(timestamp=0,
+                                       timezone=None,
+                                       committer="Foo Bar <foo@example.com>",
+                                       message="Message",
+                                       inventory_sha1=sha1,
+                                       revision_id='with_ghost')
+        rev.parent_ids = ['ghost']
+        repo.add_revision('with_ghost', rev)
+
+    def test_fetch_all_fixes_up_ghost(self):
+        # fetching from a repo with a current ghost unghosts it in referencing
+        # revisions.
+        repo = repository.Repository.open('missing_ghost')
+        rev = repo.get_revision('with_ghost')
+        from_repo = repository.Repository.open('with_ghost_rev')
+        repo.fetch(from_repo)
+        # rev must not be corrupt now
+        rev = repo.get_revision('with_ghost')
+        self.assertEqual([None, 'ghost', 'with_ghost'], repo.get_ancestry('with_ghost'))

=== modified file 'bzrlib/tests/repository_implementations/test_reconcile.py'
--- bzrlib/tests/repository_implementations/test_reconcile.py	2006-02-25 00:32:59 +0000
+++ bzrlib/tests/repository_implementations/test_reconcile.py	2006-02-28 01:13:01 +0000
@@ -19,7 +19,7 @@
 
 import bzrlib
 import bzrlib.errors as errors
-from bzrlib.reconcile import reconcile, Reconciler
+from bzrlib.reconcile import reconcile, Reconciler, RepoReconciler
 from bzrlib.revision import Revision
 from bzrlib.tests.repository_implementations.test_repository import TestCaseWithRepository
 from bzrlib.transport import get_transport
@@ -34,9 +34,21 @@
         
         t = get_transport(self.get_url())
         # an empty inventory with no revision for testing with.
-        repo = self.make_repository('inventory_no_revision')
+        # this is referenced by 'references_missing' to let us test
+        # that all the cached data is correctly converted into ghost links
+        # and the referenced inventory still cleaned.
+        repo = self.make_repository('inventory_without_revision')
         inv = EmptyTree().inventory
         repo.add_inventory('missing', inv, [])
+        sha1 = repo.add_inventory('references_missing', inv, ['missing'])
+        rev = Revision(timestamp=0,
+                       timezone=None,
+                       committer="Foo Bar <foo@example.com>",
+                       message="Message",
+                       inventory_sha1=sha1,
+                       revision_id='references_missing')
+        rev.parent_ids = ['missing']
+        repo.add_revision('references_missing', rev)
 
         # a inventory with no parents and the revision has parents..
         # i.e. a ghost.
@@ -67,10 +79,9 @@
     def test_reweave_empty(self):
         self.make_repository('empty')
         d = bzrlib.bzrdir.BzrDir.open('empty')
-        reconciler = Reconciler(d)
+        # calling on a empty repository should do nothing
+        reconciler = RepoReconciler(d.find_repository())
         reconciler.reconcile()
-        # smoke test for reconcile appears to work too.
-        reconcile(d)
         # no inconsistent parents should have been found
         self.assertEqual(0, reconciler.inconsistent_parents)
         # and no garbage inventories
@@ -82,12 +93,43 @@
                           'inventory.backup',
                           repo.get_transaction())
 
+    def test_reweave_inventory_without_revision_reconcile(self):
+        # smoke test for the all in one ui tool
+        d = bzrlib.bzrdir.BzrDir.open('inventory_without_revision')
+        reconcile(d)
+        # now the backup should have it but not the current inventory
+        repo = d.open_repository()
+        backup = repo.control_weaves.get_weave('inventory.backup',
+                                               repo.get_transaction())
+        self.assertTrue('missing' in backup.names())
+        self.assertRaises(errors.WeaveRevisionNotPresent,
+                          repo.get_inventory, 'missing')
+
+    def test_reweave_inventory_without_revision_reconciler(self):
+        # smoke test for the all in one Reconciler class,
+        # other tests use the lower level RepoReconciler.
+        d = bzrlib.bzrdir.BzrDir.open('inventory_without_revision')
+        reconciler = Reconciler(d)
+        reconciler.reconcile()
+        # no inconsistent parents should have been found
+        self.assertEqual(1, reconciler.inconsistent_parents)
+        # and one garbage inventories
+        self.assertEqual(1, reconciler.garbage_inventories)
+        # now the backup should have it but not the current inventory
+        repo = d.open_repository()
+        backup = repo.control_weaves.get_weave('inventory.backup',
+                                               repo.get_transaction())
+        self.assertTrue('missing' in backup.names())
+        self.assertRaises(errors.WeaveRevisionNotPresent,
+                          repo.get_inventory, 'missing')
+
     def test_reweave_inventory_without_revision(self):
-        d = bzrlib.bzrdir.BzrDir.open('inventory_no_revision')
-        reconciler = Reconciler(d)
+        # actual low level test.
+        d = bzrlib.bzrdir.BzrDir.open('inventory_without_revision')
+        reconciler = RepoReconciler(d.open_repository())
         reconciler.reconcile()
         # no inconsistent parents should have been found
-        self.assertEqual(0, reconciler.inconsistent_parents)
+        self.assertEqual(1, reconciler.inconsistent_parents)
         # and one garbage inventories
         self.assertEqual(1, reconciler.garbage_inventories)
         # now the backup should have it but not the current inventory
@@ -97,10 +139,14 @@
         self.assertTrue('missing' in backup.names())
         self.assertRaises(errors.WeaveRevisionNotPresent,
                           repo.get_inventory, 'missing')
+        # and the parent list for 'references_missing' should have that
+        # revision a ghost now.
+        self.assertEqual([None, 'references_missing'],
+                         repo.get_ancestry('references_missing'))
 
     def test_reweave_inventory_preserves_a_revision_with_ghosts(self):
         d = bzrlib.bzrdir.BzrDir.open('inventory_one_ghost')
-        reconciler = Reconciler(d)
+        reconciler = RepoReconciler(d.open_repository())
         reconciler.reconcile()
         # no inconsistent parents should have been found: 
         # the lack of a parent for ghost is normal
@@ -116,7 +162,7 @@
         d = bzrlib.bzrdir.BzrDir.open('inventory_ghost_present')
         repo = d.open_repository()
         self.assertEqual([None, 'ghost'], repo.get_ancestry('ghost'))
-        reconciler = Reconciler(d)
+        reconciler = RepoReconciler(repo)
         reconciler.reconcile()
         # one inconsistent parents should have been found : the
         # available but not reference parent for ghost.

